signature MONOID =
    sig
        type t    (* Typ der Individuen *)
        val sum : t * t -> t
        val zero : t
    end

signature SEMIRING =
    sig 
        type t
        val sum : t * t -> t
        val zero : t
        val prod : t * t -> t
        val unit : t
    end

signature MATRICES =
    sig
        structure Ring : SEMIRING
        type t 
        val sum : t * t -> t
        val zero : int * int -> t
        val unit : int * int -> t
        val prod : t * t -> t
        val dimension : t -> int * int
        val sub : t * (int * int) -> Ring.t
        val map : (Ring.t -> Ring.t) -> t -> t
        val tabulate : (int * int) -> (int * int -> Ring.t) -> t
        val scale : Ring.t * t -> t
        val tran : t -> t
    end 

signature MATRIX =
    sig
        structure Ring : SEMIRING
        type t 
        val dim : int * int
        val sum : t * t -> t
        val zero : t
        val prod : t * t -> t
        val unit : t
        val sub : t * (int * int) -> Ring.t
        val map : (Ring.t -> Ring.t) -> t -> t
        val tabulate : (int * int -> Ring.t) -> t
        val scale : Ring.t * t -> t
        val tran : t -> t
    end

functor ScaleDisc (val max : int) =
   struct
      type t = int
      fun sum(x,y) = Int.max(x,y)
      val zero = 0
      fun prod(x,y) = Int.min(x,y)
      val unit = max
   end

functor Matrices (R : SEMIRING) : MATRICES = 
    struct
        structure Ring = R
        type r = Ring.t  

        type t = r vector vector 

        fun dimension (S:t) = 
            let 
                val rows = Vector.length S
                val columns = Vector.length (Vector.sub(S,0)) 
                val samelengths = 
                    Vector.foldr (fn (v,b) => b andalso 
                                              (Vector.length v = columns)) 
                                 true S
            in 
                if samelengths then (rows,columns) else raise Size 
            end

        fun sub (E:t,(i,j)) = 
            let 
                val (d1,d2) = dimension E 
            in 
                if 0 <= i andalso i < d1 andalso 0 <= j andalso j < d2 
                    then Vector.sub (Vector.sub (E,i), j)
                else raise Size
            end

        fun tabulate (rows,cols) (f : int * int -> r) =
            Vector.tabulate (rows, fn i => Vector.tabulate 
                                               (cols, fn k => f(i,k))) : t

        fun map (f : r -> r) (A : t) =
            let 
                val (n,m) = dimension (A:t)
            in
                tabulate (n,m) (fn (i,j) => f (sub (A,(i,j))))
            end

        fun zero (n,m) = tabulate (n,m) (fn (i,k) => R.zero) 
        fun unit (n,m) = 
            tabulate (n,m) (fn (i,k) => 
                               if i=k then R.unit else R.zero) 

        fun sum (S:t,T:t) : t = 
            let
                val (ns,ms) = dimension S
                val (nt,mt) = dimension T
                fun sumV (v:r vector,w:r vector) : r vector = 
                    Vector.tabulate (Vector.length v, 
                                     fn i => R.sum (Vector.sub(v,i),
                                                    Vector.sub(w,i))) 
            in
                if (ns,ms)=(nt,mt) then
                    Vector.tabulate (ns, fn i => sumV (Vector.sub (S,i),
                                                       Vector.sub (T,i)))
                else raise Size
            end

        fun prod (S:t,T:t) : t =
            let 
                val (ns,ms) = dimension S
                val (nt,mt) = dimension T
                fun prodV  (v:r vector,w:r vector) : r = 
                    let 
                        fun sm i = if i=0 then R.zero 
                                   else R.sum(sm (i-1), 
                                              R.prod (Vector.sub(v,i-1),
                                                      Vector.sub(w,i-1)))
                    in 
                        sm (Vector.length v)
                    end
                fun column (T:t,j) = Vector.map (fn r => Vector.sub (r,j)) T
            in
                if ms=nt then
                    tabulate (ns,mt) (fn (i,j) => prodV (Vector.sub(S,i), 
                                                         column(T,j)))
                else raise Size
            end

        fun scale (s:r, S:t) : t =
            tabulate (dimension S) (fn (i,j) => R.prod(s,sub(S,(i,j))))

        fun tran (S:t) =
            let val (n,m) = dimension S 
            in 
                tabulate (m,n) (fn (i,j) => sub(S,(j,i)))
            end
    end

functor Matrix (structure R:SEMIRING; val rows:int; val cols:int) : MATRIX =
    let 
        structure M = Matrices(R)
    in
        struct
            structure Ring = R : SEMIRING
            val dim = (rows,cols)
            type t = M.t
            val sum = M.sum
            val zero = M.zero dim
            val prod = M.prod
            val unit = M.unit dim
            val sub = M.sub
            val map = M.map
            val tabulate = M.tabulate dim
            val scale = M.scale
            val tran = M.tran
        end
    end
